import json

from pypif.obj import *
from pypif.util.case import keys_to_snake_case
from pypif.util.pif_encoder import PifEncoder


def dump(pif, fp, **kwargs):
    """
    Convert a single Physical Information Object, or a list of such objects, into a JSON-encoded text file.

    :param pif: Object or list of objects to serialize.
    :param fp: File-like object supporting .write() method to write the serialized object(s) to.
    :param kwargs: Any options available to json.dump().
    """
    return json.dump(pif, fp, cls=PifEncoder, **kwargs)


def dumps(pif, **kwargs):
    """
    Convert a single Physical Information Object, or a list of such objects, into a JSON-encoded string.

    :param pif: Object or list of objects to serialize.
    :param kwargs: Any options available to json.dumps().
    """
    return json.dumps(pif, cls=PifEncoder, **kwargs)


def load(fp, **kwargs):
    """
    Convert content in a JSON-encoded text file to a Physical Information Object or a list of such objects.

    :param fp: File-like object supporting .read() method to deserialize from.
    :param kwargs: Any options available to json.load().
    :return: Single object derived from :class:`.Pio` or a list of such object.
    """
    return loado(json.load(fp, **kwargs))


def loads(s, **kwargs):
    """
    Convert content in a JSON-encoded string to a Physical Information Object or a list of such objects.

    :param s: String to deserialize.
    :param kwargs: Any options available to json.loads().
    :return: Single object derived from :class:`.Pio` or a list of such object.
    """
    return loado(json.loads(s, **kwargs))


def loado(obj):
    """
    Convert a dictionary or a list of dictionaries into a single Physical Information Object or a list of such objects.

    :param obj: Dictionary or list to convert to Physical Information Objects.
    :return: Single object derived from :class:`.Pio` or a list of such object.
    """
    if isinstance(obj, list):
        return [_dict_to_pio(i) for i in obj]
    elif isinstance(obj, dict):
        return _dict_to_pio(obj)
    else:
        raise ValueError('expecting list or dictionary as outermost structure')


def _dict_to_pio(d):
    """
    Convert a single dictionary object to a Physical Information Object.

    :param d: Dictionary to convert.
    :return: Single object derived from :class:`.Pio`.
    """
    d = keys_to_snake_case(d)
    if 'category' not in d:
        raise ValueError('Dictionary does not contains a category field: ' + ', '.join(d.keys()))
    elif d['category'] == 'system':
        return System(**d)
    elif d['category'] == 'system.chemical':
        return ChemicalSystem(**d)
    elif d['category'] == 'system.chemical.alloy':
        return Alloy(**d)
    elif d['category'] == 'system.chemical.alloy.phase':  # Legacy support
        return ChemicalSystem(**d)
    raise ValueError('Dictionary does not contain a valid top-level category: ' + str(d['category']))



from pypif.obj.common.instrument import Instrument
from pypif.obj.common.pio import Pio
from pypif.obj.common.software import Software


class Method(Pio):
    """
    Information about a method used in obtaining a property value.
    """

    def __init__(self, name=None, instruments=None, software=None, tags=None, **kwargs):
        """
        Constructor.

        :param name: String with the name of the method.
        :param instruments: List of dictionaries or :class:`.Instrument` objects used in the method.
        :param software: List of dictionaries or :class:`.Software` objects used in the method.
        :param kwargs: Dictionary of fields that are not supported.
        """
        super(Method, self).__init__(tags=tags, **kwargs)
        self._name = None
        self.name = name
        self._instruments = None
        self.instruments = instruments
        self._software = None
        self.software = software

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        self._validate_type('name', name, basestring)
        self._name = name

    @property
    def instruments(self):
        return self._instruments

    @instruments.setter
    def instruments(self, instruments):
        self._validate_list_type('instruments', instruments, dict, Instrument)
        self._instruments = self._get_object(Instrument, instruments)

    @instruments.deleter
    def instruments(self):
        self._instruments = None

    @property
    def software(self):
        return self._software

    @software.setter
    def software(self, software):
        self._validate_list_type('software', software, dict, Software)
        self._software = self._get_object(Software, software)

    @software.deleter
    def software(self):
        self._software = None



import numbers
from pypif.obj.common.pio import Pio


class Scalar(Pio):
    """
    Representation of a single scalar value that could represent an absolute point, an uncertain point,
    a range of values, a minimum, or a maximum.
    """

    def __init__(self, value=None, minimum=None, inclusive_minimum=None, maximum=None, inclusive_maximum=None,
                 uncertainty=None, approximate=None, tags=None, **kwargs):
        """
        Constructor.

        :param value: String or number with the value for the point.
        :param minimum: String or number with the minimum value for the point.
        :apram inclusive_minimum: Boolean with whether the minimum is inclusive.
        :param maximum: String or number with the maximum value for the point.
        :param inclusive_maximum: Boolean with whether the maximum is inclusive.
        :param uncertainty: String or number with the  isotropic uncertainty for the point.
        :param approximate: Boolean with whether the value is approximate.
        :param tags: List of strings or numbers that are tags for this object.
        :param kwargs: Dictionary of fields that are not supported.
        """
        super(Scalar, self).__init__(tags=tags, **kwargs)
        self._value = None
        self.value = value
        self._minimum = None
        self.minimum = minimum
        self._inclusive_minimum = None
        self.inclusive_minimum = inclusive_minimum
        self._maximum = None
        self.maximum = maximum
        self._inclusive_maximum = None
        self.inclusive_maximum = inclusive_maximum
        self._uncertainty = None
        self.uncertainty = uncertainty
        self._approximate = None
        self.approximate = approximate

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._validate_type('value', value, basestring, numbers.Number)
        self._value = value

    @value.deleter
    def value(self):
        self._value = None

    @property
    def minimum(self):
        return self._minimum

    @minimum.setter
    def minimum(self, minimum):
        self._validate_type('minimum', minimum, basestring, numbers.Number)
        self._minimum = minimum

    @minimum.deleter
    def minimum(self):
        self._minimum = None

    @property
    def inclusive_minimum(self):
        return self._inclusive_minimum

    @inclusive_minimum.setter
    def inclusive_minimum(self, inclusive_minimum):
        self._validate_type('inclusive_minimum', inclusive_minimum, bool)
        self._inclusive_minimum = inclusive_minimum
    
    @inclusive_minimum.deleter
    def inclusive_minimum(self):
        self._inclusive_minimum = None

    @property
    def maximum(self):
        return self._maximum

    @maximum.setter
    def maximum(self, maximum):
        self._validate_type('maximum', maximum, basestring, numbers.Number)
        self._maximum = maximum

    @maximum.deleter
    def maximum(self):
        self._maximum = None

    @property
    def inclusive_maximum(self):
        return self._inclusive_maximum

    @inclusive_maximum.setter
    def inclusive_maximum(self, inclusive_maximum):
        self._validate_type('inclusive_maximum', inclusive_maximum, bool)
        self._inclusive_maximum = inclusive_maximum

    @inclusive_maximum.deleter
    def inclusive_maximum(self):
        self._inclusive_maximum = None

    @property
    def uncertainty(self):
        return self._uncertainty

    @uncertainty.setter
    def uncertainty(self, uncertainty):
        self._validate_type('uncertainty', uncertainty, basestring, numbers.Number)
        self._uncertainty = uncertainty

    @uncertainty.deleter
    def uncertainty(self):
        self._uncertainty = None

    @property
    def approximate(self):
        return self._approximate

    @approximate.setter
    def approximate(self, approximate):
        self._validate_type('approximate', approximate, bool)
        self._approximate = approximate

    @approximate.deleter
    def approximate(self):
        self._approximate = None



from pypif.obj.common.pio import Pio


class Name(Pio):
    """
    Representation of the first and last name of a person.
    """

    def __init__(self, title=None, given=None, family=None, suffix=None, tags=None, **kwargs):
        """
        Constructor.

        :param title: Title of the person (e.g. Prof. or Dr.)
        :param given: Given (first) name of a person.
        :param family: Family (last) name of a person.
        :param suffix: Suffix of the person (e.g. Jr. or Sr.)
        :param tags: List of strings or numbers that are tags for this object.
        :param kwargs: Dictionary of fields that are not supported.
        """
        super(Name, self).__init__(tags=tags, **kwargs)
        self._title = None
        self.title = title
        self._given = None
        self.given = given
        self._family = None
        self.family = family
        self._suffix = None
        self.suffix = suffix

    @property
    def title(self):
        return self._title

    @title.setter
    def title(self, title):
        self._validate_type('title', title, basestring)
        self._title = title

    @title.deleter
    def title(self):
        self._title = None

    @property
    def given(self):
        return self._given

    @given.setter
    def given(self, given):
        self._validate_type('given', given, basestring)
        self._given = given

    @given.deleter
    def given(self):
        self._given = None

    @property
    def family(self):
        return self._family

    @family.setter
    def family(self, family):
        self._validate_type('family', family, basestring)
        self._family = family

    @family.deleter
    def family(self):
        self._family = None

    @property
    def suffix(self):
        return self._suffix

    @suffix.setter
    def suffix(self, suffix):
        self._validate_type('suffix', suffix, basestring)
        self._suffix = suffix

    @suffix.deleter
    def suffix(self):
        self._suffix = None



from pypif.obj.common.pio import Pio


class Instrument(Pio):
    """
    Information about an instrument used to take a measurement.
    """

    def __init__(self, name=None, model=None, producer=None, url=None, tags=None, **kwargs):
        """
        Constructor.

        :param name: String with the name of the instrument.
        :param model: String with the model of the instrument.
        :param producer: String with the name of the producer of the instrument.
        :param url: URL to the instrument website.
        :param tags: List of strings or numbers that are tags for this object.
        :param kwargs: Dictionary of fields that are not supported.
        """
        super(Instrument, self).__init__(tags=tags, **kwargs)
        self._name = None
        self.name = name
        self._model = None
        self.model = model
        self._producer = None
        self.producer = producer
        self._url = None
        self.url = url

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        self._validate_type('name', name, basestring)
        self._name = name

    @name.deleter
    def name(self):
        self._name = None

    @property
    def model(self):
        return self._model

    @model.setter
    def model(self, model):
        self._validate_type('model', model, basestring)
        self._model = model

    @model.deleter
    def model(self):
        self._model = None

    @property
    def producer(self):
        return self._producer

    @producer.setter
    def producer(self, producer):
        self._validate_type('producer', producer, basestring)
        self._producer = producer

    @producer.deleter
    def producer(self):
        self._producer = None

    @property
    def url(self):
        return self._url

    @url.setter
    def url(self, url):
        self._validate_type('url', url, basestring)
        self._url = url

    @url.deleter
    def url(self):
        self._url = None



import numbers
from pypif.obj.common.pio import Pio
from pypif.obj.common.scalar import Scalar


class Value(Pio):
    """
    Information about a scalar, vector, or matrix, or a list of one of those.
    """

    def __init__(self, name=None, scalars=None, vectors=None, matrices=None, units=None, tags=None, **kwargs):
        """
        Constructor.

        :param name: String with the name of the value.
        :param scalars: One or more dictionaries, strings, numbers, or :class:`.Scalar` objects.
        :param vectors: One or more lists of dictionaries, strings, numbers, or :class:`.Scalar` objects,
                each representing a vector.
        :param matrices: One of more lists of lists of dictionaries, strings, numbers, or :class:`.Scalar` objects,
                each representing a matrix with rows as the innermost lists.
        :param units: String with the units of the value.
        :param tags: List of strings or numbers that are tags for this object.
        :param kwargs: Dictionary of fields that are not supported.
        """
        super(Value, self).__init__(tags=tags, **kwargs)
        self._name = None
        self.name = name
        self._scalars = None
        self.scalars = scalars
        self._vectors = None
        self.vectors = vectors
        self._matrices = None
        self.matrices = matrices
        self._units = None
        self.units = units

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        self._validate_type('name', name, basestring)
        self._name = name

    @name.deleter
    def name(self):
        self._name = None

    @property
    def scalars(self):
        return self._scalars

    @scalars.setter
    def scalars(self, scalars):
        self._validate_list_type('scalars', scalars, dict, basestring, numbers.Number, Scalar)
        self._scalars = self._get_object(Scalar, scalars)

    @scalars.deleter
    def scalars(self):
        self._scalars = None

    @property
    def vectors(self):
        return self._vectors

    @vectors.setter
    def vectors(self, vectors):
        self._validate_nested_list_type('vectors', vectors, 2, dict, basestring, numbers.Number, Scalar)
        self._vectors = self._get_object(Scalar, vectors)

    @vectors.deleter
    def vectors(self):
        self._vectors = None

    @property
    def matrices(self):
        return self._matrices

    @matrices.setter
    def matrices(self, matrices):
        self._validate_nested_list_type('matrices', matrices, 3, dict, basestring, numbers.Number, Scalar)
        self._matrices = self._get_object(Scalar, matrices)

    @matrices.deleter
    def matrices(self):
        self._matrices = None

    @property
    def units(self):
        return self._units

    @units.setter
    def units(self, units):
        self._validate_type('units', units, basestring)
        self._units = units

    @units.deleter
    def units(self):
        self._units = None



from pypif.obj.common.method import Method
from pypif.obj.common.rcl import Rcl
from pypif.obj.common.value import Value


class Property(Value, Rcl):
    """
    Class to store information about a property and conditions under which it exists.
    """

    def __init__(self, name=None, scalars=None, vectors=None, matrices=None, units=None, conditions=None, method=None,
                 data_type=None, references=None, contacts=None, licenses=None, tags=None, **kwargs):
        """
        Constructor.

        :param name: String with the name of the property.
        :param scalars: One or more dictionaries, strings, numbers, or :class:`.Scalar` objects.
        :param vectors: One or more lists of dictionaries, strings, numbers, or :class:`.Scalar` objects,
                each representing a vector.
        :param matrices: One of more lists of lists of dictionaries, strings, numbers, or :class:`.Scalar` objects,
                each representing a matrix with rows as the innermost lists.
        :param units: String with the units of the property.
        :param conditions: List of dictionaries or :class:`.Value` objects with the conditions at which the
                property exists.
        :param method: Dictionary or :class:`.Method` object describing the method used to get the property value.
        :param data_type: String containing "EXPERIMENTAL", "COMPUTATIONAL", or "MACHINE_LEARNING" to set the
                broad category of data.
        :param references: List of dictionaries or :class:`.Reference` objects where information about the
                property is published.
        :param contacts: List of dictionaries, strings, or :class:`.Person` objects with people to contact for
                information about the property.
        :param licenses: List of dictionaries, strings, or :class:`.License` objects with licensing information
                for the property.
        :param tags: List of strings or numbers that are tags for this object.
        :param kwargs: Dictionary of fields that are not supported.
        """
        super(Property, self).__init__(name=name, scalars=scalars, vectors=vectors, matrices=matrices,
                                       units=units, tags=tags, **kwargs)
        super(Rcl, self).__init__(references=references, contacts=contacts, licenses=licenses)
        self._conditions = None
        self.conditions = conditions
        self._method = None
        self.method = method
        self._data_type = None
        self.data_type = data_type

    @property
    def conditions(self):
        return self._conditions

    @conditions.setter
    def conditions(self, conditions):
        self._validate_list_type('conditions', conditions, dict, Value)
        self._conditions = self._get_object(Value, conditions)

    @conditions.deleter
    def conditions(self):
        self._conditions = None

    @property
    def method(self):
        return self._method

    @method.setter
    def method(self, method):
        self._validate_type('method', method, dict, Method)
        self._method = self._get_object(Method, method)

    @method.deleter
    def method(self):
        self._method = None

    @property
    def data_type(self):
        return self._data_type

    @data_type.setter
    def data_type(self, data_type):
        self._validate_type('data_type', data_type, basestring)
        self._data_type = data_type

    @data_type.deleter
    def data_type(self):
        self._data_type = None



from pypif.obj.common.license import License
from pypif.obj.common.person import Person
from pypif.obj.common.pio import Pio
from pypif.obj.common.reference import Reference


class Rcl(Pio):
    """
    Base class for any objects that contain reference, contact, and license information.
    """

    def __init__(self, references=None, contacts=None, licenses=None, tags=None, **kwargs):
        """
        Constructor.

        :param references: List of dictionaries or :class:`.Reference` objects where information about this
                item is published.
        :param contacts: List of dictionaries, strings, or :class:`.Person` objects with people that can be
                contacted for information about this item.
        :param licenses: List of dictionaries, strings, or :class:`.License` objects with licensing information
                for this item.
        :param tags: List of strings or numbers that are tags for this object.
        :param kwargs: Dictionary of fields that are not supported.
        """
        super(Rcl, self).__init__(tags=tags, **kwargs)
        self._references = None
        self.references = references
        self._contacts = None
        self.contacts = contacts
        self._licenses = None
        self.licenses = licenses

    @property
    def references(self):
        return self._references

    @references.setter
    def references(self, references):
        self._validate_list_type('references', references, dict, Reference)
        self._references = self._get_object(Reference, references)

    @references.deleter
    def references(self):
        self._references = None

    @property
    def contacts(self):
        return self._contacts

    @contacts.setter
    def contacts(self, contacts):
        self._validate_list_type('contacts', contacts, dict, basestring, Person)
        self._contacts = self._get_object(Person, contacts)

    @contacts.deleter
    def contacts(self):
        self._contacts = None

    @property
    def licenses(self):
        return self._licenses

    @licenses.setter
    def licenses(self, licenses):
        self._validate_list_type('licenses', licenses, dict, basestring, License)
        self._licenses = self._get_object(License, licenses)

    @licenses.deleter
    def licenses(self):
        self._licenses = None



from id import Id
from instrument import Instrument
from license import License
from method import Method
from name import Name
from pages import Pages
from person import Person
from process_step import ProcessStep
from property import Property
from reference import Reference
from scalar import Scalar
from source import Source
from software import Software
from value import Value



from pypif.obj.system.system import System
from pypif.obj.system.chemical.common.composition import Composition


class ChemicalSystem(System):
    """
    Representation of a general chemical system.
    """

    def __init__(self, uid=None, names=None, ids=None, source=None, quantity=None, chemical_formula=None,
                 composition=None, properties=None, preparation=None, sub_systems=None, references=None,
                 contacts=None, licenses=None, tags=None, **kwargs):
        """
        Constructor.

        :param uid: String with the permanent ID for this record.
        :param names: List of strings with common names of the chemical system.
        :param ids: List of dictionaries, strings, numbers, or :class:`.Id` objects that identify the system.
        :param source: Dictionary, string, or :class:`.Source` object with the source of the system.
        :param quantity: Dictionary or :class:`.Quantity` object with the quantity of the system.
        :param chemical_formula: String with the chemical formula.
        :param composition: List of dictionaries or :class:`.Composition` objects that describe the composition of
                the chemical system.
        :param properties: List of dictionaries or :class:`.Property` objects with properties of the system.
        :param preparation: List of dictionaries or :class:`.ProcessStep` objects with the preparation
                information of the system.
        :param sub_systems: List of dictionaries or :class:`.System` objects with the subsystems of the system.
        :param references: List of dictionaries or :class:`.Reference` objects where information about the system
                is published.
        :param contacts: List of dictionaries, strings, or :class:`.Person` objects with people to contact for
                information about the system.
        :param licenses: List of dictionaries, strings, or :class:`.License` objects with licensing information
                for data about the system.
        :param tags: List of strings or numbers that are tags for this object.
        :param kwargs: Dictionary of fields that are not supported.
        """
        super(ChemicalSystem, self).__init__(uid=uid, names=names, ids=ids, source=source, quantity=quantity,
                                             properties=properties, preparation=preparation, sub_systems=sub_systems,
                                             references=references, contacts=contacts, licenses=licenses, tags=tags,
                                             **kwargs)
        self._chemical_formula = None
        self.chemical_formula = chemical_formula
        self._composition = None
        self.composition = composition
        self.category = kwargs['category'] if 'category' in kwargs else 'system.chemical'

    @property
    def chemical_formula(self):
        return self._chemical_formula

    @chemical_formula.setter
    def chemical_formula(self, chemical_formula):
        self._validate_type('chemical_formula', chemical_formula, basestring)
        self._chemical_formula = chemical_formula

    @chemical_formula.deleter
    def chemical_formula(self):
        self._chemical_formula = None

    @property
    def composition(self):
        return self._composition

    @composition.setter
    def composition(self, composition):
        self._validate_list_type('composition', composition, dict, Composition)
        self._composition = self._get_object(Composition, composition)

    @composition.deleter
    def composition(self):
        self._composition = None



from setuptools import setup, find_packages

setup(name='pypif',
      version='1.0.13',
      url='http://github.com/CitrineInformatics/pypif',
      description='Python tools for working with the Physical Information File (PIF)',
      author='Kyle Michel',
      author_email='kyle@citrine.io',
      packages=find_packages())



from pypif.obj.common.pio import Pio


class Pages(Pio):
    """
    Representation of the starting and ending pages of a reference.
    """

    def __init__(self, start=None, end=None, tags=None, **kwargs):
        """
        Constructor.

        :param start: String or integer with the starting page.
        :param end: String or integer with the ending page.
        :param tags: List of strings or numbers that are tags for this object.
        :param kwargs: Dictionary of fields that are not supported.
        """
        super(Pages, self).__init__(tags=tags, **kwargs)
        self._start = None
        self.start = start
        self._end = None
        self.end = end

    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, start):
        self._validate_type('start', start, basestring, int)
        self._start = start

    @start.deleter
    def start(self):
        self._start = None

    @property
    def end(self):
        return self._end

    @end.setter
    def end(self, end):
        self._validate_type('end', end, basestring, int)
        self._end = end

    @end.deleter
    def end(self):
        self._end = None
